/**
 * Java FsdlRequest Parser
 *
 * @version		Alpha 0.3
 *
 * @copyright	1999-2018 OP3FT
 *
 * This package is provided by the OP3FT to help you create your own
 * dynamic Frogans sites.

 * You are free to use, copy, distribute, modify, adapt, or share
 * this file under the terms set forth in the
 * COPYRIGHT-AND-LICENSE.txt file in this archive.
 *
 * For information about the Java FsdlRequest Parser, see the
 * README.txt file in this archive.
 *
 * @file		FsdlRequest.java
 *
 */

package fsdl_request_parser;


/**
 * FsdlRequest
 *
 * Class used to parse an FSDL-Request document sent by Frogans Player
 * to the server hosting a Frogans site.
 *
 **/

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class FsdlRequest
{
	/**
	 * Maximum size of FSDL-Request documents, in bytes.
	 */
	public static final String FSDL_REQUEST_DOCUMENT_VERSION = "3.0";

	/**
	 * Maximum size of FSDL-Request documents, in bytes.
	 */
	public static final int DOCUMENT_SIZE_MAX = 65536;


	/**
	 * Maximum number of session fields
	 */
	public static final int SESSION_FIELDS_COUNT_MAX = 16;


	/**
	 * Maximum number of file fields
	 */
	public static final int FILE_FIELDS_COUNT_MAX = 16;


	/**
	 * Maximum number of entry fields
	 */
	public static final int ENTRY_FIELDS_COUNT_MAX = 8;


	/**
	 * Maximum number of characters for the key attribute value
	 */
	public static final int KEY_ATTRIBUTE_VALUE_SIZE_MAX = 24;



	/**
	* FsdlRequest constructor
	*/
	public FsdlRequest()
	{
		doClear();
	}


	/**
	* Clears an instance
	* @return void
	*/
	public void clear()
	{
		doClear();
	}


	/**
	 * Initializes an instance
	 *
	 * @param String fsdlRequestDocumentVersion		Possible values are defined in FsdlRequestDocumentVersion
	 * @param InputStream receivedDocument 			FSDL-Request document received by the server;
	 * 												Document size, in bytes, must be less than or equal to DOCUMENT_SIZE_MAX;
	 * 												Encoding can be UTF-8 or UTF-16 (reflecting the encoding for the Frogans site)
	 * @return boolean
	 */
	public boolean initialize(String fsdlRequestDocumentVersion, InputStream receivedDocument) {
		return doInitialize(fsdlRequestDocumentVersion, receivedDocument);
	}


	/**
	 * Gets the error message
	 * Available if the initialization has failed
	 *
	 * @return String
	 */
	public String getErrorMessage() {
		return errorMessage;
	}


	/**
	 * Gets fsdlRequestWanted
	 * Instance must be initialized.
	 *
	 * Always applicable.
	 *
	 * @return Value is one of the following constants:
	 * 		FsdlRequestWanted.FSDL_DOCUMENT
	 * 		FsdlRequestWanted.AUXILIARY_FILE
	 *		(FSDL_REQUEST_UNDEFINED if not initialized)
	 */
	public int getFsdlRequestWanted()
	{
		if (!initialized)
		{
			return FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		}

		return fsdlRequestWanted;
	}


	/**
	 * Gets fsdlRequestNavigation
	 * Instance must be initialized.
	 *
	 * Applicable only if fsdlRequestWanted is FsdlRequestWanted.SDL_DOCUMENT.
	 *
	 * @return Value is one of the following constants:
	 * 		FsdlRequestNavigation.BUTTON
	 * 		FsdlRequestNavigation.NEXT
	 * 		FsdlRequestNavigation.REDIRECT
	 * 		FsdlRequestNavigation.OPEN
	 *		(FSDL_REQUEST_UNDEFINED if not initialized or not applicable)
	 */
	public int getFsdlRequestNavigation()
	{
		if (!initialized)
		{
			return FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		}

		return fsdlRequestNavigation;
	}


	/**
	 * Gets fsdlRequestKind
	 * Instance must be initialized.
	 *
	 * Applicable only if fsdlRequestWanted is FsdlRequestWanted.AUXILIARY_FILE.
	 *
	 * @return Value is one of the following constants:
	 * 		FsdlRequestKind.IMAGE
		 *		(FSDL_REQUEST_UNDEFINED if not initialized or not applicable)
	 */
	public int getFsdlRequestKind()
	{
		if (!initialized)
		{
			return FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		}

		return fsdlRequestKind;
	}


	/**
	 * Gets sessionFields
	 * Instance must be initialized.
	 *
	 * Always applicable.
	 *
	 * @return Value is an array of up to FsdlRequest.SESSION_FIELDS_COUNT_MAX key-value pairs
	 * 		   The array can be empty.
	 */
	public HashMap<String,String> getSessionFields()
	{
		if (!initialized)
		{
			return new HashMap<String,String>();
		}

		return sessionFields;
	}


	/**
	 * Gets fileFields
	 * Instance must be initialized.
	 *
	 * Applicable only if fsdlRequestWanted is FsdlRequestWanted.FSDL_DOCUMENT
	 * 						and fsdlRequestNavigation is not FsdlRequestNavigation.OPEN
	 * 				   or
	 * 				   if fsdlRequestWanted is FsdlRequestWanted.AUXILIARY_FILE
	 *
	 * @return Value is an array of up to FsdlRequest.FILE_FIELDS_COUNT_MAX key-value pairs
	 * 		   The array can be empty.
	 */
	public HashMap<String,String> getFileFields()
	{
		if (!initialized)
		{
			return new HashMap<String,String>();
		}

		return fileFields;
	}


	/**
	 * Gets entryFields
	 * Instance must be initialized.
	 *
	 * Applicable only if fsdlRequestWanted is FsdlRequestWanted.FSDL_DOCUMENT
	 * 						and fsdlRequestNavigation is FsdlRequestNavigation.BUTTON
	 *
	 * @return Value is an array of up to FsdlRequest.ENTRY_FIELDS_COUNT_MAX key-value pairs
	 * 		   The array can be empty.
	 */
	public HashMap<String,String> getEntryFields()
	{
		if (initialized == false)
		{
			return new HashMap<String,String>();
		}

		return entryFields;
	}


	/* ========================================================== */
	/* ====== The properties and methods below are private ====== */
	/* ========================================================== */


	private boolean initialized;
	private int fsdlRequestWanted;
	private int fsdlRequestNavigation;
	private int fsdlRequestKind;

	private HashMap<String, String> sessionFields;
	private HashMap<String, String> fileFields;
	private HashMap<String, String> entryFields;

	private String errorMessage;

	private void doClear()
	{
		initialized = false;
		fsdlRequestWanted = FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		fsdlRequestNavigation = FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		fsdlRequestKind = FsdlRequestUndefined.FSDL_REQUEST_UNDEFINED;
		sessionFields = new HashMap<String,String>();
		fileFields = new HashMap<String,String>();
		entryFields = new HashMap<String,String>();
		errorMessage = "";
	}

	private void appendLine(String line)
	{
		if (line.trim() != "" && line != null)  // ignore empty lines or lines containing spaces only
		{
			errorMessage += "Error: " + line;
			if (!line.endsWith("\n"))  // add an EOL character if line does not have one
			{
				errorMessage += "\n";
			}
		}
	}

	private boolean doInitialize(String fsdlRequestDocumentVersion, InputStream receivedDocument)
	{
		doClear();

		// Check parameters

		if (!fsdlRequestDocumentVersion.equals(FSDL_REQUEST_DOCUMENT_VERSION))
		{
			appendLine("fsdlRequestDocumentVersion parameter value is not supported in this version of the FsdlRequest parser.");
			return false;
		}

		// Convert the inputStream to a String

		final char[] buffer = new char[1024];
		final StringBuilder sb = new StringBuilder();
		Reader rd = null;
		try {
			rd = new InputStreamReader(receivedDocument, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			appendLine("[InputStream to String conversion error] " + e.getLocalizedMessage());
		}
		int rsz = 0;
		for(;;) {
			try {
				rsz = rd.read(buffer, 0, buffer.length);
			} catch (IOException e) {
				appendLine("[InputStream to String conversion error] " + e.getLocalizedMessage());
			}
			if (rsz < 0)
				break;
			sb.append(buffer, 0, rsz);
		}
		String receivedDocumentStr = sb.toString().substring(1);

		// Check the FSDL-Request document size

		if (receivedDocumentStr.length() > DOCUMENT_SIZE_MAX)
		{
			appendLine("The document size is greater than the limit of " + DOCUMENT_SIZE_MAX + " bytes.");
			return false;
		}

		// Trims XML leading and trailing white spaces, newlines and tabs

		receivedDocumentStr = receivedDocumentStr.trim().replaceAll("\\n", "").replaceAll("\\t", "");

		if (receivedDocumentStr.replaceAll("\\s", "").equals(""))
		{
			appendLine("The document is empty or contains white spaces only.");
			return false;
		}


		// Parse the FSDL-Request document in XML

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		Element root = null;

        try {
        	DocumentBuilder builder = factory.newDocumentBuilder();
		    Document document= builder.parse(new InputSource(new StringReader(receivedDocumentStr)));
		    root = document.getDocumentElement();
        }
        catch (final ParserConfigurationException e) {
        	appendLine("[XML parsing error] " + e.getLocalizedMessage());
			return false;
		}
        catch (final SAXException e) {
        	appendLine("[XML parsing error] " + e.getLocalizedMessage());
			return false;
        }
        catch (final IOException e) {
        	appendLine("[XML parsing error] " + e.getLocalizedMessage());
			return false;
        }

        // TODO: raise error if XML declaration is not included in the document (XML parser does not fail currently)


		// Validate the FSDL-Request document

		// Validate the root element (<frogans-fsdl-request>)

		if (!root.getNodeName().equals("frogans-fsdl-request"))
		{
			appendLine("The name of the root element must be <frogans-fsdl-request>.");
			return false;
		}

		NamedNodeMap rootAttributes = root.getAttributes();

		if (rootAttributes.getLength() != 1)
		{
			appendLine("The <frogans-fsdl-request> element must have exactly 1 attribute.");
			return false;
		}

		Node rootAttributeVersion = rootAttributes.getNamedItem("version");

		if (rootAttributeVersion == null)
		{
			appendLine("The name of the attribute of the <frogans-fsdl-request> element must be \"version\".");
			return false;
		}

		if (!rootAttributeVersion.getNodeValue().equals(FSDL_REQUEST_DOCUMENT_VERSION))
		{
			appendLine("The value of the \"version\" attribute of the <frogans-fsdl-request> element must be '3.0'.");
			return false;
		}

		NodeList rootChildren = root.getChildNodes();

		if (rootChildren.getLength() != 4)
		{
			appendLine("The <frogans-fsdl-request> element must contain exactly 4 child elements.");
			return false;
		}

		Node rootChild1 = rootChildren.item(0);
		Node rootChild2 = rootChildren.item(1);
		Node rootChild3 = rootChildren.item(2);
		Node rootChild4 = rootChildren.item(3);

		// Validate the first child element (<request>)

		if (!rootChild1.getNodeName().equals("request"))
		{
			appendLine("The name of the first child element of the <frogans-fsdl-request> element must be \"request\".");
			return false;
		}

		NamedNodeMap rootChild1Attributes = rootChild1.getAttributes();

		if (rootChild1Attributes.getLength() != 2)
		{
			appendLine("The <request> element must have exactly 2 attributes.");
			return false;
		}

		Node rootChild1AttributeWanted = rootChild1Attributes.getNamedItem("wanted");

		if (rootChild1AttributeWanted == null)
		{
			appendLine("The <request> element must have a \"wanted\" attribute.");
			return false;
		}

		if (!(rootChild1AttributeWanted.getNodeValue().equals("fsdl-document")) &&
			!(rootChild1AttributeWanted.getNodeValue().equals("auxiliary-file")))
		{
			appendLine("The value of the \"wanted\" attribute of the <frogans-fsdl-request> element must be 'fsdl-document' or 'auxiliary-file'.");
			return false;
		}

		Node rootChild1AttributeNavigation = null;

		if (rootChild1AttributeWanted.getNodeValue().equals("fsdl-document"))
		{
			rootChild1AttributeNavigation = rootChild1Attributes.getNamedItem("navigation");

			if (rootChild1AttributeNavigation == null)
			{
				appendLine("The <request> element must have a \"navigation\" attribute, since the value of its \"wanted\" attribute is 'fsdl-document'.");
				return false;
			}

			if (!(rootChild1AttributeNavigation.getNodeValue().equals("button")) &&
				!(rootChild1AttributeNavigation.getNodeValue().equals("next")) &&
				!(rootChild1AttributeNavigation.getNodeValue().equals("redirect")) &&
				!(rootChild1AttributeNavigation.getNodeValue().equals("open")))
			{
				appendLine("The value of the \"navigation\" attribute of the <request> element must be 'button', 'next', 'redirect', or 'open'.");
				return false;
			}
		}
		else  // rootChild1AttributeWanted.getNodeValue().equals("auxiliary-file")
		{
			Node rootChild1AttributeKind = rootChild1Attributes.getNamedItem("kind");

			if (rootChild1AttributeKind == null)
			{
				appendLine("The <request> element must have a \"kind\" attribute, since the value of its \"wanted\" attribute is 'auxiliary-file'.");
				return false;
			}

			if (!rootChild1AttributeKind.getNodeValue().equals("image"))
			{
				appendLine("The value of the \"kind\" attribute of the <request> element must be 'image'.");
				return false;
			}
		}

		NodeList rootChild1Children = rootChild1.getChildNodes();

		if (rootChild1Children.getLength() != 0)
		{
			appendLine("The <request> element must not contain any child element.");
			return false;
		}

		// Validate the second child element (<session-fields>)

		boolean canContainFields = true;

		if (!validateField(rootChild2,
							  	"session-fields",
							  	"second",
							  	canContainFields,
							  	SESSION_FIELDS_COUNT_MAX))
		{
			return false;
		}

		// Validate the third child element (<file-fields>)

		canContainFields = false;

		if (rootChild1AttributeWanted.getNodeValue().equals("fsdl-document"))
		{
			if (!rootChild1AttributeNavigation.getNodeValue().equals("open"))
			{
				canContainFields = true;
			}
		}
		else // rootChild1AttributeWanted.getNodeValue().equals("auxiliary-file")
		{
			canContainFields = true;
		}

		if (!validateField(rootChild3,
								"file-fields",
								"third",
								canContainFields,
								FILE_FIELDS_COUNT_MAX))
		{
			return false;
		}

		// Validate the fourth child element (<entry-fields>)

		canContainFields = false;

		if (rootChild1AttributeWanted.getNodeValue().equals("fsdl-document"))
		{
			if (rootChild1AttributeNavigation.getNodeValue().equals("button"))
			{
				canContainFields = true;
			}
		}

		if (!validateField(rootChild4,
								"entry-fields",
								"fourth",
								canContainFields,
								ENTRY_FIELDS_COUNT_MAX))
		{
			return false;
		}


		// TODO: validate unicity of key value within child elements, or across the FSDL-Request document (if confirmed by FSDL specification)


		// Validation is successful, so we can now fill in the private properties

		initialized = true;

		if (rootChild1AttributeWanted.getNodeValue().equals("fsdl-document"))
		{
			fsdlRequestWanted = FsdlRequestWanted.FSDL_DOCUMENT;

			if (rootChild1AttributeNavigation.getNodeValue().equals("button"))
			{
				fsdlRequestNavigation = FsdlRequestNavigation.BUTTON;
			}
			else if (rootChild1AttributeNavigation.getNodeValue().equals("next"))
			{
				fsdlRequestNavigation = FsdlRequestNavigation.NEXT;
			}
			else if (rootChild1AttributeNavigation.getNodeValue().equals("redirect"))
			{
				fsdlRequestNavigation = FsdlRequestNavigation.REDIRECT;
			}
			else  // rootChild1AttributeNavigation.getNodeValue().equals("open")
			{
				fsdlRequestNavigation = FsdlRequestNavigation.OPEN;
			}
		}
		else  // rootChild1AttributeWanted.equals("auxiliary-file")
		{
			fsdlRequestWanted = FsdlRequestWanted.AUXILIARY_FILE;

			// rootChild1AttributeKind.equals("image")

			fsdlRequestKind = FsdlRequestKind.IMAGE;
		}

		Node field;
		Node fieldAttribute;

		NodeList rootChild2Children = rootChild2.getChildNodes();
		if (rootChild2Children.getLength() > 0)
		{
			for (int i=0; i<rootChild2Children.getLength(); i++)
			{
				field = rootChild2Children.item(i);
				fieldAttribute = field.getAttributes().getNamedItem("key");
				this.sessionFields.put(fieldAttribute.getNodeValue(), field.getTextContent());
			}
		}

		NodeList rootChild3Children = rootChild3.getChildNodes();
		if (rootChild3Children.getLength() > 0)
		{
			for (int i=0; i<rootChild3Children.getLength(); i++)
			{
				field = rootChild3Children.item(i);
				fieldAttribute = field.getAttributes().getNamedItem("key");
				this.fileFields.put(fieldAttribute.getNodeValue(), field.getTextContent());
			}
		}

		NodeList rootChild4Children = rootChild4.getChildNodes();
		if (rootChild4Children.getLength() > 0)
		{
			for (int i=0; i<rootChild4Children.getLength(); i++)
			{
				field = rootChild4Children.item(i);
				fieldAttribute = field.getAttributes().getNamedItem("key");
				this.entryFields.put(fieldAttribute.getNodeValue(), field.getTextContent());
			}
		}

		return true;
	}


	private boolean validateField(Node field,
									String elementName,
									String elementOrder,
									boolean canContainFields,
									int fieldCountMax)
	{
		if (!field.getNodeName().equals(elementName))
		{
			appendLine("The name of the " + elementOrder + " child element of the <frogans-fsdl-request> element must be \"" + elementName + "\".");
			return false;
		}

		NamedNodeMap fieldAttributes = field.getAttributes();

		if (fieldAttributes.getLength() != 0)
		{
			appendLine("The <" + elementName + "> element must not have any attribute.");
			return false;
		}

		NodeList fieldChildren = field.getChildNodes();

		int fieldChildrenCount = fieldChildren.getLength();

		if ((fieldChildrenCount > 0) && (!canContainFields))
		{
			appendLine("The <" + elementName + "> element must not contain any child element given the value of the attributes of the <request> element.");
			return false;
		}

		if (fieldChildrenCount > fieldCountMax)
		{
			appendLine("The <" + elementName + "> element must not contain more than " + fieldCountMax + " child elements.");
			return false;
		}

		if (fieldChildrenCount > 0)
		{
			Node fieldAttributeKey;
			for (int i=0; i<fieldChildrenCount; i++)
			{
				field = fieldChildren.item(i);
				if (!field.getNodeName().equals("field"))
				{
					appendLine("The name of the child elements of the <" + elementName + "> element must be \"field\".");
					return false;
				}

				fieldAttributes = field.getAttributes();

				if (fieldAttributes.getLength() != 1)
				{
					appendLine("The <field> elements of the <" + elementName + "> element must have exactly 1 attribute.");
					return false;
				}

				fieldAttributeKey = fieldAttributes.item(0);

				if (!fieldAttributeKey.getNodeName().equals("key"))
				{
					appendLine("The name of the attribute of the <field> elements of the <" + elementName + "> element must be \"key\".");
					return false;
				}

				// Syntax of key attribute value Validation

				if (fieldAttributeKey.getNodeValue().length() > KEY_ATTRIBUTE_VALUE_SIZE_MAX) {
					appendLine("The value of the \"key\" attribute of the <field> elements of the <" + elementName + "> element exceeds " + KEY_ATTRIBUTE_VALUE_SIZE_MAX + " characters.");
					return false;
				}

				Pattern p = Pattern.compile("[^a-zA-Z0-9_-]");
				boolean hasInvalidCharacter = p.matcher(fieldAttributeKey.getNodeValue()).find();
				if (hasInvalidCharacter) {
					appendLine("The value of the \"key\" attribute of the <field> elements of the <" + elementName + "> element is invalid (at least one character value is not allowed.");
					return false;
				}
			}
		}

		return true;
	}
}
